@@ -19,10 +19,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include "pango-break.h"
 #include "pango-modules.h"
+#include "pango-script-private.h"
 #include <string.h>
 
 #define PARAGRAPH_SEPARATOR 0x2029
@@ -345,7 +346,7 @@ static const int line_break_indexes[] = {
 };
 
 #define BREAK_TYPE_SAFE(btype)            \
-	 (btype < G_N_ELEMENTS(line_break_indexes) ? btype : G_UNICODE_BREAK_UNKNOWN)
+	 ((btype) < G_N_ELEMENTS(line_break_indexes) ? (btype) : G_UNICODE_BREAK_UNKNOWN)
 #define BREAK_INDEX(btype)                \
 	 (line_break_indexes[(btype)])
 #define BREAK_ROW(before_type)            \
@@ -353,7 +354,7 @@ static const int line_break_indexes[] = {
 #define BREAK_OP(before_type, after_type) \
 	 (BREAK_ROW (before_type)[BREAK_INDEX (after_type)])
 #define IN_BREAK_TABLE(btype)             \
-	 (btype < G_N_ELEMENTS(line_break_indexes) && BREAK_INDEX(btype) < INDEX_END_OF_TABLE)
+	 ((btype) < G_N_ELEMENTS(line_break_indexes) && BREAK_INDEX((btype)) < INDEX_END_OF_TABLE)
 
 
 
@@ -417,27 +418,6 @@ static const CharJamoProps HangulJamoProps[] = {
 #define JAMO_TYPE(btype)      \
 	(IS_JAMO(btype) ? (btype - G_UNICODE_BREAK_HANGUL_L_JAMO) : NO_JAMO)
 
-
-
-
-/* "virama script" is just an optimization; it includes a bunch of
- * scripts without viramas in them
- */
-#define VIRAMA_SCRIPT(wc)        ((wc) >= 0x0901 && (wc) <= 0x17FF)
-#define VIRAMA(wc) ((wc) == 0x094D || \
-		    (wc) == 0x09CD || \
-		    (wc) == 0x0A4D || \
-		    (wc) == 0x0ACD || \
-		    (wc) == 0x0B4D || \
-		    (wc) == 0x0BCD || \
-		    (wc) == 0x0C4D || \
-		    (wc) == 0x0CCD || \
-		    (wc) == 0x0D4D || \
-		    (wc) == 0x0DCA || \
-		    (wc) == 0x0E3A || \
-		    (wc) == 0x0F84 || \
-		    (wc) == 0x1039 || \
-		    (wc) == 0x17D2)
 /* Types of Japanese characters */
 #define JAPANESE(wc) ((wc) >= 0x2F00 && (wc) <= 0x30FF)
 #define KANJI(wc)    ((wc) >= 0x2F00 && (wc) <= 0x2FDF)
@@ -451,7 +431,6 @@ static const CharJamoProps HangulJamoProps[] = {
 #define HANGUL(wc) ((wc) >= 0xAC00 && (wc) <= 0xD7A3)
 #define BACKSPACE_DELETES_CHARACTER(wc) (!LATIN (wc) && !CYRILLIC (wc) && !GREEK (wc) && !KANA(wc) && !HANGUL(wc))
 
-
 /* p. 132-133 of Unicode spec table 5-6 will help understand this */
 typedef enum
 {
@@ -503,9 +482,9 @@ typedef enum
 void
 pango_default_break (const gchar   *text,
 		     gint           length,
-		     PangoAnalysis *analysis,
+		     PangoAnalysis *analysis G_GNUC_UNUSED,
 		     PangoLogAttr  *attrs,
-		     int            attrs_len)
+		     int            attrs_len G_GNUC_UNUSED)
 {
   /* The rationale for all this is in section 5.15 of the Unicode 3.0 book,
    * the line breaking stuff is also in TR14 on unicode.org
@@ -532,6 +511,35 @@ pango_default_break (const gchar   *text,
   GUnicodeBreakType prev_break_type; /* skips spaces */
   gboolean prev_was_break_space;
 
+  /* See Grapheme_Cluster_Break Property Values table of UAX#29 */
+  typedef enum
+  {
+    GB_Other,
+    GB_ControlCRLF,
+    GB_Extend,
+    GB_Prepend,
+    GB_SpacingMark,
+    GB_InHangulSyllable, /* Handles all of L, V, T, LV, LVT rules */
+  } GraphemeBreakType;
+  GraphemeBreakType prev_GB_type = GB_Other;
+
+  /* See Word_Break Property Values table of UAX#29 */
+  typedef enum
+  {
+    WB_Other,
+    WB_NewlineCRLF,
+    WB_ExtendFormat,
+    WB_Katakana,
+    WB_ALetter,
+    WB_MidNumLet,
+    WB_MidLetter,
+    WB_MidNum,
+    WB_Numeric,
+    WB_ExtendNumLet,
+  } WordBreakType;
+  WordBreakType prev_prev_WB_type = WB_Other, prev_WB_type = WB_Other;
+  gint prev_WB_i = -1;
+
   WordType current_word_type = WordNone;
   gunichar last_word_letter = 0;
   gunichar base_character = 0;
@@ -551,14 +559,17 @@ pango_default_break (const gchar   *text,
 
   next = text;
 
-  prev_type = (GUnicodeType) -1;
+  prev_type = G_UNICODE_PARAGRAPH_SEPARATOR;
   prev_break_type = G_UNICODE_BREAK_UNKNOWN;
   prev_was_break_space = FALSE;
   prev_wc = 0;
   prev_jamo = NO_JAMO;
 
   if (length == 0 || *text == '\0')
-    next_wc = PARAGRAPH_SEPARATOR;
+    {
+      next_wc = PARAGRAPH_SEPARATOR;
+      almost_done = TRUE;
+    }
   else
     next_wc = g_utf8_get_char (next);
 
@@ -574,6 +585,11 @@ pango_default_break (const gchar   *text,
       JamoType jamo;
       gboolean makes_hangul_syllable;
 
+      /* UAX#29 boundaries */
+      gboolean is_grapheme_boundary;
+      gboolean is_word_boundary;
+
+
       wc = next_wc;
       break_type = next_break_type;
 
@@ -631,109 +647,267 @@ pango_default_break (const gchar   *text,
        */
       attrs[i].is_expandable_space = (0x0020 == wc || 0x00A0 == wc);
 
-      /* ---- Cursor position breaks (Grapheme breaks) ---- */
+      /* ---- UAX#29 Grapheme Boundaries ---- */
+      {
+	GraphemeBreakType GB_type;
+        /* Find the GraphemeBreakType of wc */
+	GB_type = GB_Other;
+	switch ((int) type)
+	  {
+	  case G_UNICODE_FORMAT:
+	    if (wc == 0x200C && wc == 0x200D)
+	      {
+		GB_type = GB_Extend; /* U+200C and U+200D are Other_Grapheme_Extend */
+		break;
+	      }
+	    /* fall through */
+	  case G_UNICODE_CONTROL:
+	  case G_UNICODE_LINE_SEPARATOR:
+	  case G_UNICODE_PARAGRAPH_SEPARATOR:
+	    GB_type = GB_ControlCRLF;
+	    break;
 
-      if (wc == '\n')
-	{
-	  /* Break before line feed unless prev char is a CR */
+	  case G_UNICODE_OTHER_LETTER:
+	    if (makes_hangul_syllable)
+	      GB_type = GB_InHangulSyllable;
+	    else if ((wc & 0x0E00) == 0x0E00)
+	      {
+	        /* Thai and Lao stuff hardcoded in UAX#29 */
+		if ((wc >= 0x0E40 && wc <= 0x0E44) || (wc >= 0x0EC0 && wc <= 0x0EC4))
+		  GB_type = GB_Prepend; /* Prepend */
+		else if (wc == 0x0E30 || wc == 0x0E32 || wc == 0x0E33 || wc == 0x0E45 ||
+			 wc == 0x0EB0 || wc == 0x0EB2 || wc == 0x0EB3)
+		  GB_type = GB_Extend; /* Exceptions in the Extend definition */
+	      }
+	    break;
 
-	  if (prev_wc != '\r')
-	    attrs[i].is_cursor_position = TRUE;
-	  else
-	    attrs[i].is_cursor_position = FALSE;
-	}
-      else if (i == 0 ||
-	       prev_type == G_UNICODE_CONTROL ||
-	       prev_type == G_UNICODE_FORMAT)
-	{
-	  /* Break at first position (must be special cased, or if the
-	   * first char is say a combining mark there won't be a
-	   * cursor position at the start, which seems wrong to me
-	   * ???? - maybe it makes sense though, who knows)
-	   */
-	  /* break after all format or control characters */
-	  attrs[i].is_cursor_position = TRUE;
-	}
-      else
-	{
-	  switch (type)
-	    {
-	    case G_UNICODE_CONTROL:
-	    case G_UNICODE_FORMAT:
-	      /* Break before all format or control characters */
-	      attrs[i].is_cursor_position = TRUE;
-	      break;
+	  case G_UNICODE_MODIFIER_LETTER:
+	    if (wc >= 0xFF9E && wc <= 0xFF9F)
+	      GB_type = GB_Extend; /* Other_Grapheme_Extend */
+	    break;
 
-	    case G_UNICODE_COMBINING_MARK:
-	    case G_UNICODE_ENCLOSING_MARK:
-	    case G_UNICODE_NON_SPACING_MARK:
-	      /* Unicode spec includes "Combining marks plus Tibetan
-	       * subjoined characters" as joining chars, but lists the
-	       * Tibetan subjoined characters as combining marks, and
-	       * g_unichar_type() returns NON_SPACING_MARK for the Tibetan
-	       * subjoined characters. So who knows, beats me.
-	       */
+	  case G_UNICODE_COMBINING_MARK:
+	    GB_type = GB_SpacingMark; /* SpacingMark */
+	    if (wc >= 0x0900)
+	      {
+		if (wc == 0x09BE || wc == 0x09D7 ||
+		    wc == 0x0B3E || wc == 0x0B57 || wc == 0x0BBE || wc == 0x0BD7 ||
+		    wc == 0x0CC2 || wc == 0x0CD5 || wc == 0x0CD6 ||
+		    wc == 0x0D3E || wc == 0x0D57 || wc == 0x0DCF || wc == 0x0DDF ||
+		    wc == 0x1D165 || (wc >= 0x1D16E && wc <= 0x1D172))
+		  GB_type = GB_Extend; /* Other_Grapheme_Extend */
+	      }
+	    break;
 
-	      /* It's a joining character, break only if preceded by
-	       * control or format; we already handled the case where
-	       * it was preceded earlier, so here we know it wasn't,
-	       * don't break
-	       */
-	      attrs[i].is_cursor_position = FALSE;
-	      break;
+	  case G_UNICODE_ENCLOSING_MARK:
+	  case G_UNICODE_NON_SPACING_MARK:
+	    GB_type = GB_Extend; /* Grapheme_Extend */
+	    break;
+	  }
 
-	    case G_UNICODE_LOWERCASE_LETTER:
-	    case G_UNICODE_MODIFIER_LETTER:
-	    case G_UNICODE_OTHER_LETTER:
-	    case G_UNICODE_TITLECASE_LETTER:
-	    case G_UNICODE_UPPERCASE_LETTER:
+	/* Grapheme Cluster Boundary Rules */
+	/* We apply Rules GB1 and GB2 at the end of the function */
+	if (wc == '\n' && prev_wc == '\r')
+	  is_grapheme_boundary = FALSE; /* Rule GB3 */
+	else if (prev_GB_type == GB_ControlCRLF || GB_type == GB_ControlCRLF)
+	  is_grapheme_boundary = TRUE; /* Rules GB4 and GB5 */
+	else if (GB_type == GB_InHangulSyllable)
+	  is_grapheme_boundary = FALSE; /* Rules GB6, GB7, GB8 */
+	else if (GB_type == GB_Extend)
+	  is_grapheme_boundary = FALSE; /* Rule GB9 */
+	else if (GB_type == GB_SpacingMark)
+	  is_grapheme_boundary = FALSE; /* Rule GB9a */
+	else if (prev_GB_type == GB_Prepend)
+	  is_grapheme_boundary = FALSE; /* Rule GB9b */
+	else
+	  is_grapheme_boundary = TRUE;  /* Rule GB10 */
 
-	      if (makes_hangul_syllable)
-		attrs[i].is_cursor_position = FALSE;
-	      else
-		{
-		  /* Handle non-Hangul-syllable non-combining chars */
+	prev_GB_type = GB_type;
 
-		  /* Break before Jamo if they are in a broken sequence or
-		   * next to non-Jamo; break if preceded by Jamo; don't
-		   * break if a letter is preceded by a virama; break in
-		   * all other cases. No need to check whether we are or are
-		   * preceded by Jamo explicitly, since a Jamo is not
-		   * a virama, we just break in all cases where we
-		   * aren't a or preceded by a virama.  Don't fool with
-		   * viramas if we aren't part of a script that uses them.
-		   */
+	attrs[i].is_cursor_position = is_grapheme_boundary;
+	/* If this is a grapheme boundary, we have to decide if backspace
+	 * deletes a character or the whole grapheme cluster */
+	if (is_grapheme_boundary)
+	  attrs[i].backspace_deletes_character = BACKSPACE_DELETES_CHARACTER (base_character);
+	else
+	  attrs[i].backspace_deletes_character = FALSE;
+      }
 
-		  if (VIRAMA_SCRIPT (wc))
-		    {
-		      /* Check whether we're preceded by a virama; this
-		       * could use some optimization.
-		       */
-		      if (VIRAMA (prev_wc))
-			attrs[i].is_cursor_position = FALSE;
-		      else
-			attrs[i].is_cursor_position = TRUE;
-		    }
-		  else
-		    {
-		      attrs[i].is_cursor_position = TRUE;
-		    }
+      /* ---- UAX#29 Word Boundaries ---- */
+      {
+	is_word_boundary = FALSE;
+	if (is_grapheme_boundary) /* Rules WB3 and WB4 */
+	  {
+	    PangoScript script;
+	    WordBreakType WB_type;
+
+	    script = pango_script_for_unichar (wc);
+
+	    /* Find the WordBreakType of wc */
+	    WB_type = WB_Other;
+
+	    if (script == PANGO_SCRIPT_KATAKANA)
+	      WB_type = WB_Katakana;
+
+	    if (WB_type == WB_Other)
+	      switch (wc >> 8)
+	        {
+		case 0x30:
+		  if (wc == 0x3031 || wc == 0x3032 || wc == 0x3033 || wc == 0x3034 || wc == 0x3035 ||
+		      wc == 0x309b || wc == 0x309c || wc == 0x30a0 || wc == 0x30fc)
+		    WB_type = WB_Katakana; /* Katakana exceptions */
+		  break;
+		case 0xFF:
+		  if (wc == 0xFF70)
+		    WB_type = WB_Katakana; /* Katakana exceptions */
+		  else if (wc >= 0xFF9E || wc <= 0xFF9F)
+		    WB_type = WB_ExtendFormat; /* Other_Grapheme_Extend */
+		  break;
+		case 0x05:
+		  if (wc == 0x05F3)
+		    WB_type = WB_ALetter; /* ALetter exceptions */
+		  break;
 		}
-	      break;
 
-	    default:
-	      /* Some weirdo char, just break here, why not */
-	      attrs[i].is_cursor_position = TRUE;
-	      break;
-	    }
-	}
+	    if (WB_type == WB_Other)
+	      switch ((int) break_type)
+	        {
+		case G_UNICODE_BREAK_NUMERIC:
+		  if (wc != 0x066C)
+		    WB_type = WB_Numeric; /* Numeric */
+		  break;
+		case G_UNICODE_BREAK_INFIX_SEPARATOR:
+		  if (wc != 0x003A && wc != 0xFE13 && wc != 0x002E)
+		    WB_type = WB_MidNum; /* MidNum */
+		  break;
+		}
+
+	    if (WB_type == WB_Other)
+	      switch ((int) type)
+		{
+		case G_UNICODE_CONTROL:
+		  if (wc != 0x000D && wc != 0x000A && wc != 0x000B && wc != 0x000C && wc != 0x0085)
+		    break;
+		  /* fall through */
+		case G_UNICODE_LINE_SEPARATOR:
+		case G_UNICODE_PARAGRAPH_SEPARATOR:
+		  WB_type = WB_NewlineCRLF; /* CR, LF, Newline */
+		  break;
+
+		case G_UNICODE_FORMAT:
+		case G_UNICODE_COMBINING_MARK:
+		case G_UNICODE_ENCLOSING_MARK:
+		case G_UNICODE_NON_SPACING_MARK:
+		  WB_type = WB_ExtendFormat; /* Extend, Format */
+		  break;
+
+		case G_UNICODE_CONNECT_PUNCTUATION:
+		  WB_type = WB_ExtendNumLet; /* ExtendNumLet */
+		  break;
+
+		case G_UNICODE_INITIAL_PUNCTUATION:
+		case G_UNICODE_FINAL_PUNCTUATION:
+		  if (wc == 0x2018 || wc == 0x2019)
+		    WB_type = WB_MidNumLet; /* MidNumLet */
+		  break;
+		case G_UNICODE_OTHER_PUNCTUATION:
+		  if (wc == 0x0027 || wc == 0x002e || wc == 0x2024 ||
+		      wc == 0xfe52 || wc == 0xff07 || wc == 0xff0e)
+		    WB_type = WB_MidNumLet; /* MidNumLet */
+		  else if (wc == 0x00b7 || wc == 0x05f4 || wc == 0x2027 || wc == 0x003a || wc == 0x0387 ||
+			   wc == 0xfe13 || wc == 0xfe55 || wc == 0xff1a)
+		    WB_type = WB_MidLetter; /* WB_MidLetter */
+		  else if (wc == 0x066c ||
+			   wc == 0xfe50 || wc == 0xfe54 || wc == 0xff0c || wc == 0xff1b)
+		    WB_type = WB_MidNum; /* MidNum */
+		  break;
+
+		case G_UNICODE_OTHER_SYMBOL:
+		  if (wc >= 0x24B6 && wc <= 0x24E9) /* Other_Alphabetic */
+		    goto Alphabetic;
+		  break;
+
+		case G_UNICODE_OTHER_LETTER:
+		case G_UNICODE_LETTER_NUMBER:
+		  if (wc == 0x3006 || wc == 0x3007 ||
+		      (wc >= 0x3021 && wc <= 0x3029) ||
+		      (wc >= 0x3038 && wc <= 0x303A) ||
+		      (wc >= 0x3400 && wc <= 0x4DB5) ||
+		      (wc >= 0x4E00 && wc <= 0x9FC3) ||
+		      (wc >= 0xF900 && wc <= 0xFA2D) ||
+		      (wc >= 0xFA30 && wc <= 0xFA6A) ||
+		      (wc >= 0xFA70 && wc <= 0xFAD9) ||
+		      (wc >= 0x20000 && wc <= 0x2A6D6) ||
+		      (wc >= 0x2F800 && wc <= 0x2FA1D))
+		    break; /* ALetter exceptions: Ideographic */
+		  goto Alphabetic;
+
+		case G_UNICODE_LOWERCASE_LETTER:
+		case G_UNICODE_MODIFIER_LETTER:
+		case G_UNICODE_TITLECASE_LETTER:
+		case G_UNICODE_UPPERCASE_LETTER:
+		Alphabetic:
+		  if (break_type != G_UNICODE_BREAK_COMPLEX_CONTEXT && script != PANGO_SCRIPT_HIRAGANA)
+		    WB_type = WB_ALetter; /* ALetter */
+		  break;
+		}
+
+	    /* Grapheme Cluster Boundary Rules */
+
+	    /* We apply Rules WB1 and WB2 at the end of the function */
+
+	    if (prev_wc == 0x3031 && wc == 0x41)
+	      g_debug ("Y %d %d", prev_WB_type, WB_type);
+	    if (prev_WB_type == WB_NewlineCRLF && prev_WB_i + 1 == i)
+	      {
+	        /* The extra check for prev_WB_i is to correctly handle sequences like
+		 * Newline ÷ Extend × Extend
+		 * since we have not skipped ExtendFormat yet.
+		 */
+	        is_word_boundary = TRUE; /* Rule WB3a */
+	      }
+	    else if (WB_type == WB_NewlineCRLF)
+	      is_word_boundary = TRUE; /* Rule WB3b */
+	    else if (WB_type == WB_ExtendFormat)
+	      is_word_boundary = FALSE; /* Rules WB4? */
+	    else if ((prev_WB_type == WB_ALetter  ||
+		      prev_WB_type == WB_Numeric  ||
+		      prev_WB_type == WB_ExtendNumLet) &&
+		     (     WB_type == WB_ALetter  ||
+		           WB_type == WB_Numeric  ||
+		           WB_type == WB_ExtendNumLet))
+	      is_word_boundary = FALSE; /* Rules WB5, WB8, WB9, WB10, WB13a, WB13b */
+	    else if ((prev_WB_type == WB_Katakana ||
+		      prev_WB_type == WB_ExtendNumLet) &&
+		     (     WB_type == WB_Katakana ||
+		           WB_type == WB_ExtendNumLet))
+	      is_word_boundary = FALSE; /* Rules WB13, WB13a, WB13b */
+	    else if ((prev_prev_WB_type == WB_ALetter && WB_type == WB_ALetter) &&
+		     (prev_WB_type == WB_MidLetter || prev_WB_type == WB_MidNumLet))
+	      {
+		attrs[prev_WB_i].is_word_boundary = FALSE; /* Rule WB6 */
+		is_word_boundary = FALSE; /* Rule WB7 */
+	      }
+	    else if ((prev_prev_WB_type == WB_Numeric && WB_type == WB_Numeric) &&
+		     (prev_WB_type == WB_MidNum || prev_WB_type == WB_MidNumLet))
+	      {
+		is_word_boundary = FALSE; /* Rule WB11 */
+		attrs[prev_WB_i].is_word_boundary = FALSE; /* Rule WB12 */
+	      }
+	    else
+	      is_word_boundary = TRUE; /* Rule WB14 */
+
+	    if (WB_type != WB_ExtendFormat)
+	      {
+		prev_prev_WB_type = prev_WB_type;
+		prev_WB_type = WB_type;
+		prev_WB_i = i;
+	      }
+	  }
+
+	attrs[i].is_word_boundary = is_word_boundary;
+      }
 
-      /* If this is a grapheme boundary, we have to decide if backspace
-       * deletes a character or the whole grapheme cluster */
-      if (attrs[i].is_cursor_position)
-	attrs[i].backspace_deletes_character = BACKSPACE_DELETES_CHARACTER (base_character);
-      else
-	attrs[i].backspace_deletes_character = FALSE;
 
       /* ---- Line breaking ---- */
 
@@ -762,7 +936,7 @@ pango_default_break (const gchar   *text,
 	  attrs[i].is_char_break = TRUE;
 
 	  /* Make any necessary replacements first */
-	  switch (prev_break_type)
+	  switch ((int) prev_break_type)
 	    {
 	    case G_UNICODE_BREAK_HANGUL_L_JAMO:
 	    case G_UNICODE_BREAK_HANGUL_V_JAMO:
@@ -793,7 +967,7 @@ pango_default_break (const gchar   *text,
 	      ;
 	    }
 
-	  switch (prev_break_type)
+	  switch ((int) prev_break_type)
 	    {
 	    case G_UNICODE_BREAK_MANDATORY:
 	    case G_UNICODE_BREAK_LINE_FEED:
@@ -835,7 +1009,7 @@ pango_default_break (const gchar   *text,
 	       * boundaries.
 	       */
 
-	      switch (break_type)
+	      switch ((int) break_type)
 		{
 		case G_UNICODE_BREAK_MANDATORY:
 		case G_UNICODE_BREAK_LINE_FEED:
@@ -900,29 +1074,29 @@ pango_default_break (const gchar   *text,
 	      break;
 	    }
 
-	  if (break_op != BREAK_ALREADY_HANDLED)
+	  switch (break_op)
 	    {
-	      switch (break_op)
-		{
-		case BREAK_PROHIBITED:
-		  /* can't break here */
-		  attrs[i].is_char_break = FALSE;
-		  break;
+	    case BREAK_PROHIBITED:
+	      /* can't break here */
+	      attrs[i].is_char_break = FALSE;
+	      break;
 
-		case BREAK_IF_SPACES:
-		  /* break if prev char was space */
-		  if (prev_was_break_space)
-		    attrs[i].is_line_break = TRUE;
-		  break;
+	    case BREAK_IF_SPACES:
+	      /* break if prev char was space */
+	      if (prev_was_break_space)
+		attrs[i].is_line_break = TRUE;
+	      break;
 
-		case BREAK_ALLOWED:
-		  attrs[i].is_line_break = TRUE;
-		  break;
+	    case BREAK_ALLOWED:
+	      attrs[i].is_line_break = TRUE;
+	      break;
 
-		default:
-		  g_assert_not_reached ();
-		  break;
-		}
+	    case BREAK_ALREADY_HANDLED:
+	      break;
+
+	    default:
+	      g_assert_not_reached ();
+	      break;
 	    }
 	}
 
@@ -944,7 +1118,7 @@ pango_default_break (const gchar   *text,
       if (current_word_type != WordNone)
 	{
 	  /* Check for a word end */
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_COMBINING_MARK:
 	    case G_UNICODE_ENCLOSING_MARK:
@@ -1011,7 +1185,7 @@ pango_default_break (const gchar   *text,
       else
 	{
 	  /* Check for a word start */
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_LOWERCASE_LETTER:
 	    case G_UNICODE_MODIFIER_LETTER:
@@ -1053,7 +1227,7 @@ pango_default_break (const gchar   *text,
        */
 
 #define MAYBE_START_NEW_SENTENCE                                \
-	      switch (type)                                     \
+	      switch ((int) type)                               \
 		{                                               \
 		case G_UNICODE_LINE_SEPARATOR:                  \
 		case G_UNICODE_PARAGRAPH_SEPARATOR:             \
@@ -1086,7 +1260,7 @@ pango_default_break (const gchar   *text,
       /* Break after line/para separators except carriage return
        * followed by newline
        */
-      switch (prev_type)
+      switch ((int) prev_type)
 	{
 	case G_UNICODE_LINE_SEPARATOR:
 	case G_UNICODE_PARAGRAPH_SEPARATOR:
@@ -1108,7 +1282,7 @@ pango_default_break (const gchar   *text,
       /* break before para/line separators except newline following
        * carriage return
        */
-      switch (type)
+      switch ((int) type)
 	{
 	case G_UNICODE_LINE_SEPARATOR:
 	case G_UNICODE_PARAGRAPH_SEPARATOR:
@@ -1131,7 +1305,7 @@ pango_default_break (const gchar   *text,
 	{
 	case STATE_SENTENCE_OUTSIDE:
 	  /* Start sentence if we have non-whitespace/format/control */
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_LINE_SEPARATOR:
 	    case G_UNICODE_PARAGRAPH_SEPARATOR:
@@ -1169,7 +1343,7 @@ pango_default_break (const gchar   *text,
 	   * loosely-specified OTHER_PUNCTUATION such as period,
 	   * comma, etc.; follow Unicode rules for breaks
 	   */
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_OTHER_PUNCTUATION:
 	    case G_UNICODE_CLOSE_PUNCTUATION:
@@ -1213,7 +1387,7 @@ pango_default_break (const gchar   *text,
 	  /* End sentence on anything besides more punctuation; follow
 	   * rules for breaks
 	   */
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_OTHER_PUNCTUATION:
 	    case G_UNICODE_CLOSE_PUNCTUATION:
@@ -1267,7 +1441,7 @@ pango_default_break (const gchar   *text,
 	   * we had to see a space, which ends the sentence.
 	   */
 
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_SPACE_SEPARATOR:
 	      /* continue in this state */
@@ -1306,7 +1480,7 @@ pango_default_break (const gchar   *text,
 	  break;
 
 	case STATE_SENTENCE_DOT:
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_CLOSE_PUNCTUATION:
 	      sentence_state = STATE_SENTENCE_POST_DOT_CLOSE;
@@ -1335,7 +1509,7 @@ pango_default_break (const gchar   *text,
 	  break;
 
 	case STATE_SENTENCE_POST_DOT_CLOSE:
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_SPACE_SEPARATOR:
 	      possible_sentence_end = i;
@@ -1363,7 +1537,7 @@ pango_default_break (const gchar   *text,
 
 	  possible_sentence_boundary = i;
 
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_SPACE_SEPARATOR:
 	      /* remain in current state */
@@ -1399,7 +1573,7 @@ pango_default_break (const gchar   *text,
 	  break;
 
 	case STATE_SENTENCE_POST_DOT_OPEN:
-	  switch (type)
+	  switch ((int) type)
 	    {
 	    case G_UNICODE_OPEN_PUNCTUATION:
 	      /* continue in current state */
@@ -1466,6 +1640,17 @@ pango_default_break (const gchar   *text,
 	  type != G_UNICODE_NON_SPACING_MARK)
 	base_character = wc;
     }
+  i--;
+
+  attrs[i].is_cursor_position = TRUE;  /* Rule GB2 */
+  attrs[0].is_cursor_position = TRUE;  /* Rule GB1 */
+
+  attrs[i].is_word_boundary = TRUE;  /* Rule WB2 */
+  attrs[0].is_word_boundary = TRUE;  /* Rule WB1 */
+
+  attrs[i].is_line_break = TRUE;  /* Rule LB3 */
+  attrs[0].is_line_break = FALSE; /* Rule LB2 */
+
 }
 
 static gboolean
@@ -1679,8 +1864,8 @@ pango_get_log_attrs (const char    *text,
   PangoEngineLang *range_engine;
   static guint engine_type_id = 0;
   static guint render_type_id = 0;
-  PangoAnalysis analysis = { 0 };
-  PangoScriptIter *iter;
+  PangoAnalysis analysis = { NULL };
+  PangoScriptIter iter;
 
   g_return_if_fail (length == 0 || text != NULL);
   g_return_if_fail (log_attrs != NULL);
@@ -1699,17 +1884,17 @@ pango_get_log_attrs (const char    *text,
 
   chars_broken = 0;
 
-  iter = pango_script_iter_new (text, length);
-  pango_script_iter_get_range (iter, &range_start, &range_end, &script);
+  _pango_script_iter_init (&iter, text, length);
+  pango_script_iter_get_range (&iter, &range_start, &range_end, &script);
   range_engine = (PangoEngineLang*) pango_map_get_engine (lang_map, script);
   g_assert (range_start == text);
 
-  while (pango_script_iter_next (iter))
+  while (pango_script_iter_next (&iter))
     {
       const char *run_start, *run_end;
       PangoEngineLang* run_engine;
 
-      pango_script_iter_get_range (iter, &run_start, &run_end, &script);
+      pango_script_iter_get_range (&iter, &run_start, &run_end, &script);
       run_engine = (PangoEngineLang*) pango_map_get_engine (lang_map, script);
       g_assert (range_end == run_start);
 
@@ -1725,13 +1910,13 @@ pango_get_log_attrs (const char    *text,
 	}
       range_end = run_end;
     }
-  pango_script_iter_free (iter);
+  _pango_script_iter_fini (&iter);
 
   g_assert (length < 0 || range_end == text + length);
 
   chars_broken += tailor_segment (range_start, range_end, range_engine, chars_broken, &analysis, log_attrs);
 
-  if (chars_broken + 1 < attrs_len)
+  if (chars_broken + 1 > attrs_len)
     g_warning ("pango_get_log_attrs: attrs_len should have been at least %d, but was %d.  Expect corrupted memory.",
 	       chars_broken + 1,
 	       attrs_len);
