@@ -19,7 +19,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 #include <glib.h>
 #include "pango-glyph.h"
 #include "pango-font.h"
@@ -61,14 +61,28 @@ pango_glyph_string_set_size (PangoGlyphString *string, gint new_len)
   while (new_len > string->space)
     {
       if (string->space == 0)
-	string->space = 1;
-      else
-	string->space *= 2;
-
-      if (string->space < 0)
 	{
-	  g_warning ("glyph string length overflows maximum integer size, truncated");
-	  new_len = string->space = G_MAXINT - 8;
+	  string->space = 4;
+	}
+      else
+	{
+	  const guint max_space =
+	    MIN (G_MAXINT, G_MAXSIZE / MAX (sizeof(PangoGlyphInfo), sizeof(gint)));
+
+	  guint more_space = (guint)string->space * 2;
+
+	  if (more_space > max_space)
+	    {
+	      more_space = max_space;
+
+	      if ((guint)new_len > max_space)
+		{
+		  g_error ("%s: failed to allocate glyph string of length %i\n",
+			   G_STRLOC, new_len);
+		}
+	    }
+
+	  string->space = more_space;
 	}
     }
 
@@ -92,17 +106,23 @@ pango_glyph_string_get_type (void)
 
 /**
  * pango_glyph_string_copy:
- * @string: a #PangoGlyphString.
+ * @string: a #PangoGlyphString, may be %NULL
  *
- *  Copy a glyph string and associated storage.
+ * Copy a glyph string and associated storage.
  *
  * Return value: the newly allocated #PangoGlyphString, which
- *               should be freed with pango_glyph_string_free().
+ *               should be freed with pango_glyph_string_free(),
+ *               or %NULL if @string was %NULL.
  */
 PangoGlyphString *
 pango_glyph_string_copy (PangoGlyphString *string)
 {
-  PangoGlyphString *new_string = g_slice_new (PangoGlyphString);
+  PangoGlyphString *new_string;
+
+  if (string == NULL)
+    return NULL;
+  
+  new_string = g_slice_new (PangoGlyphString);
 
   *new_string = *string;
 
@@ -116,13 +136,16 @@ pango_glyph_string_copy (PangoGlyphString *string)
 
 /**
  * pango_glyph_string_free:
- * @string:    a #PangoGlyphString.
+ * @string: a #PangoGlyphString, may be %NULL
  *
  * Free a glyph string and associated storage.
  */
 void
 pango_glyph_string_free (PangoGlyphString *string)
 {
+  if (string == NULL)
+    return;
+
   g_free (string->glyphs);
   g_free (string->log_clusters);
   g_slice_free (PangoGlyphString, string);
@@ -312,49 +335,47 @@ pango_glyph_string_get_logical_widths (PangoGlyphString *glyphs,
 				       int               embedding_level,
 				       int              *logical_widths)
 {
-  int i, j;
-  int last_cluster = 0;
-  int width = 0;
-  int last_cluster_width = 0;
-  const char *p = text;		/* Points to start of current cluster */
+  /* Build a PangoGlyphItem so we can use PangoGlyphItemIter.
+   * This API should have been made to take a PangoGlyphItem... */
+  PangoItem item = {0, length, g_utf8_strlen (text, length),
+		    {NULL, NULL, NULL,
+		     embedding_level, PANGO_GRAVITY_AUTO, 0,
+		     PANGO_SCRIPT_UNKNOWN, NULL,
+		     NULL}};
+  PangoGlyphItem glyph_item = {&item, glyphs};
+  PangoGlyphItemIter iter;
+  gboolean has_cluster;
+  int dir;
 
-  for (i=0; i<=glyphs->num_glyphs; i++)
+  dir = embedding_level % 2 == 0 ? +1 : -1;
+  for (has_cluster = pango_glyph_item_iter_init_start (&iter, &glyph_item, text);
+       has_cluster;
+       has_cluster = pango_glyph_item_iter_next_cluster (&iter))
     {
-      int glyph_index = (embedding_level % 2 == 0) ? i : glyphs->num_glyphs - i - 1;
+      int glyph_index, char_index, num_chars, cluster_width = 0, char_width;
 
-      /* If this glyph belongs to a new cluster, or we're at the end, find
-       * the start of the next cluster, and assign the widths for this cluster.
-       */
-      if (i == glyphs->num_glyphs || p != text + glyphs->log_clusters[glyph_index])
-	{
-	  int next_cluster = last_cluster;
-
-	  if (i < glyphs->num_glyphs)
-	    {
-	      while (p < text + glyphs->log_clusters[glyph_index])
-		{
-		  next_cluster++;
-		  p = g_utf8_next_char (p);
-		}
-	    }
-	  else
-	    {
-	      while (p < text + length)
-		{
-		  next_cluster++;
-		  p = g_utf8_next_char (p);
-		}
-	    }
-
-	  for (j = last_cluster; j < next_cluster; j++)
-	    logical_widths[j] = (width - last_cluster_width) / (next_cluster - last_cluster);
-
-	  last_cluster = next_cluster;
-	  last_cluster_width = width;
+      for (glyph_index  = iter.start_glyph;
+	   glyph_index != iter.end_glyph;
+	   glyph_index += dir)
+        {
+	  cluster_width += glyphs->glyphs[glyph_index].geometry.width;
 	}
 
-      if (i < glyphs->num_glyphs)
-	width += glyphs->glyphs[glyph_index].geometry.width;
+      num_chars = iter.end_char - iter.start_char;
+      if (num_chars) /* pedantic */
+        {
+	  char_width = cluster_width / num_chars;
+
+	  for (char_index = iter.start_char;
+	       char_index < iter.end_char;
+	       char_index++)
+	    {
+	      logical_widths[char_index] = char_width;
+	    }
+
+	  /* add any residues to the first char */
+	  logical_widths[iter.start_char] += cluster_width - (char_width * num_chars);
+	}
     }
 }
 
@@ -485,6 +506,12 @@ pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
   if (trailing)
     cluster_offset += 1;
 
+  if (G_UNLIKELY (!cluster_chars)) /* pedantic */
+    {
+      *x_pos = start_xpos;
+      return;
+    }
+
   *x_pos = ((cluster_chars - cluster_offset) * start_xpos +
 	    cluster_offset * end_xpos) / cluster_chars;
 }
@@ -495,7 +522,7 @@ pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
  * @text:      the text for the run
  * @length:    the number of bytes (not characters) in text.
  * @analysis:  the analysis information return from pango_itemize()
- * @x_pos:     the x offset (in #PangoGlyphUnit)
+ * @x_pos:     the x offset (in Pango units)
  * @index_:    location to store calculated byte index within @text
  * @trailing:  location to store a boolean indicating
  *             whether the user clicked on the leading or trailing
